#include "common.h"
#include "debug_io.h"
#include "layout.h"
#include "write_typed_value.h"
#include <string.h>

#define debug(fun_name, msg) pio_debug (__FILE__ "::" fun_name, msg)

static void w_value_type (pio_buffer *ob, vtd_value_type t) {
  debug ("w_value_type", "start writing");
  vt_to_ob (ob, t);
  debug ("w_value_type", "finish writing");
}

static void w_layout (pio_buffer *ob, layout l) {
  layout_to_ob (ob, l);
}

static void w_size (pio_buffer *ob, size_t siz) {
  buf_printf (ob, "<%i>\n", siz);
}

static void w_size_matrix (pio_buffer *ob, size_t siz1, size_t siz2) {
  buf_printf (ob, "<%i, %i>\n", siz1, siz2);
}

static void string_to_ascii (pio_buffer *ob, const char *s) {
  unsigned char c;

  while ((c = *s++)) {
    if (c > 31 && c < 127)
      buf_add_char (ob, c);
    else {
      buf_add_char (ob, '\\');
      if (c == '\\' || c == '"')
	buf_add_char (ob, c);
      else {
	buf_printf (ob, "%03hhu", c);
      }
    }
  }
}

static void output_ascii_string (pio_buffer *ob, const char *s) {
  buf_add_char (ob, '"');
  string_to_ascii (ob, s);
  buf_add_char (ob, '"');
}

static void w_bool (pio_buffer *ob, bool b) {
  buf_add_string (ob, b ? "true" : "false");
}

static void w_string (pio_buffer *ob, struct string str) {
  w_size (ob, str.ssize);
  output_ascii_string (ob, str.str);
}

static void w_int (pio_buffer *ob, int i) {
  buf_printf (ob, "%i", i);
}

static void w_native_int (pio_buffer *ob, natint i) {
  /* must know what a natint is */
  buf_printf (ob, "%li", i);		/* is this OK if int != long ? */
}

static void w_int32 (pio_buffer *ob, int32 i) {
  buf_printf (ob, "%li", i);
}

static void w_int64 (pio_buffer *ob, int64 i) {
  buf_printf (ob, "%lli", i);
}

static void w_float (pio_buffer *ob, double f) {
  if (binary_mode) {
    buf_add_string (ob, "&<8>");
    union { double d; int64 i64; } tem;
    tem.d = f;
    int64 l = tem.i64;
    unsigned char s[8];
    unsigned int i = 0;
    while (i < sizeof (s)) {
      s[i++] = l;
      l >>= 8;
    }
    buf_add_string_w_len (ob, (char *)&f, 8);
  } else
    buf_printf (ob, "%23.16e", f);
}

/* You get whatever polymorphism you can... */
#define w_couple(elt_type, couple_type, elt_writer, fieldnam)		\
  static void w_##elt_type##_couple (pio_buffer *ob, couple_type couple) { \
    buf_add_char (ob, '(');						\
    elt_writer (ob, couple.fieldnam##1);				\
    buf_add_string (ob, ", ");						\
    elt_writer (ob, couple.fieldnam##2);				\
    buf_add_char (ob, ')');						\
  }

#define w_triple(elt_type, triple_type, elt_writer, fieldnam)		\
  static void w_##elt_type##_triple (pio_buffer *ob, triple_type triple) { \
    buf_add_char (ob, '(');						\
    elt_writer (ob, triple.fieldnam##1);				\
    buf_add_string (ob, ", ");						\
    elt_writer (ob, triple.fieldnam##2);				\
    buf_add_string (ob, ", ");						\
    elt_writer (ob, triple.fieldnam##3);				\
    buf_add_char (ob, ')');						\
  }

w_couple (bool, struct cbool, w_bool, b)
w_triple (bool, struct tbool, w_bool, b)
w_couple (int, struct cint, w_int, i)
w_triple (int, struct tint, w_int, i)
w_couple (native_int, struct cnat, w_native_int, nat)
w_triple (native_int, struct tnat, w_native_int, nat)
w_couple (int32, struct cint32, w_int32, i32)
w_triple (int32, struct tint32, w_int32, i32)
w_couple (int64, struct cint64, w_int64, i64)
w_triple (int64, struct tint64, w_int64, i64)
w_couple (float, struct cfloat, w_float, f)
w_triple (float, struct tfloat, w_float, f)

#define w_item(w_lexem, ob, lexem)		\
  w_lexem (ob, (lexem));			\
  buf_add_string (ob, ";\n")

#define w_items(w_lexem, ob, v, siz) {		\
    size_t __i_ = 0;				\
    for (__i_ = 0; __i_ < siz; __i_++) {	\
      w_item (w_lexem, ob, (v)[__i_]);		\
    }						\
  }

#define w_vector(w_lexem, ob, v, siz) {		\
    buf_add_string (ob, "[1\n");		\
    w_size (ob, siz);				\
    w_items (w_lexem, ob, v, siz);		\
    buf_add_string (ob, "1]");			\
  }

/* Only for doubles; w_lexem = w_int has a different calling sequence */
static void w_line_float_items (layout layout, /* void w_lexem(pio_buffer *ob, double) */ pio_buffer *ob, struct matrix *m) {
  check_layout (2, layout); /* already done by w_layout in w_matrix (but it's cheap) */

  size_t i = 0, j = 0, size_i = m->msize1, size_j = m->msize2;
#define index_FORTRAN(i, j) (((j) * size_i) + (i))
#define index_C(i, j) (((i) * size_j) + (j))
  for (i = 0; i < size_i; i++) {
    buf_add_string (ob, "[|\n");
    for (j = 0; j < size_j; j++) {
      size_t idx = (layout == LAYOUT_C) ? index_C (i, j) : index_FORTRAN (i, j);
      w_item (w_float, ob, m->mf[idx]);
    }
    buf_add_string (ob, "|];\n");
  }
} /* advienne que pourra */

#define w_matrix(layout, /* w_lexem, */ ob, m) { 	\
    buf_add_string (ob, "[2\n");			\
    w_size_matrix (ob, (m)->msize1, (m)->msize2);	\
    w_layout (ob, layout);				\
    buf_add_char (ob, '\n');				\
    w_line_float_items (layout, /*w_lexem, */ ob, m);	\
    buf_add_string (ob, "2]");				\
  }

static void w_value (pio_buffer *ob, TV_t *v) {
  switch (v->tv_tag) {
  /* w_scalar */
  case TV_Sbool:		w_bool			(ob, v->v.b);		break;
  case TV_Sstring:		w_string		(ob, v->v.str);		break;
  case TV_Sint:			w_int			(ob, v->v.i);		break;
  case TV_Snative_int:		w_native_int		(ob, v->v.nat);		break;
  case TV_Sint32:		w_int32			(ob, v->v.i32);		break;
  case TV_Sint64:		w_int64			(ob, v->v.i64);		break;
  case TV_Sfloat:		w_float			(ob, v->v.f);		break;
  case TV_Sbool_couple:		w_bool_couple		(ob, v->v.cb);		break;
  case TV_Sint_couple:		w_int_couple		(ob, v->v.ci);		break;
  case TV_Snative_int_couple:	w_native_int_couple	(ob, v->v.cnat);	break;
  case TV_Sint32_couple:	w_int32_couple		(ob, v->v.ci32);	break;
  case TV_Sint64_couple:	w_int64_couple		(ob, v->v.ci64);	break;
  case TV_Sfloat_couple:	w_float_couple		(ob, v->v.cf);		break;
  case TV_Sbool_triple:		w_bool_triple		(ob, v->v.tb);		break;
  case TV_Sint_triple:		w_int_triple		(ob, v->v.ti);		break;
  case TV_Snative_int_triple:	w_native_int_triple	(ob, v->v.tnat);	break;
  case TV_Sint32_triple:	w_int32_triple		(ob, v->v.ti32);	break;
  case TV_Sint64_triple:	w_int64_triple		(ob, v->v.ti64);	break;
  case TV_Sfloat_triple:	w_float_triple		(ob, v->v.tf);		break;
    /* w_vect */
  case TV_Vbool:		w_vector (w_bool,		ob, v->v.vec.p.vb, v->v.vec.vsize);	break;
  case TV_Vstring:		w_vector (w_string,		ob, v->v.vec.p.vs, v->v.vec.vsize);	break;
  case TV_Vint:			w_vector (w_int,		ob, v->v.vec.p.vi, v->v.vec.vsize);	break;
  case TV_Vnative_int:		w_vector (w_native_int,		ob, v->v.vec.p.vnat, v->v.vec.vsize);	break;
  case TV_Vint32:		w_vector (w_int32,		ob, v->v.vec.p.vi32, v->v.vec.vsize);	break;
  case TV_Vint64:		w_vector (w_int64,		ob, v->v.vec.p.vi64, v->v.vec.vsize);	break;
  case TV_Vfloat:		w_vector (w_float,		ob, v->v.vec.p.vf, v->v.vec.vsize);	break;
  case TV_Vbool_couple:		w_vector (w_bool_couple,	ob, v->v.vec.p.vcb, v->v.vec.vsize);	break;
  case TV_Vint_couple:		w_vector (w_int_couple,		ob, v->v.vec.p.vci, v->v.vec.vsize);	break;
  case TV_Vnative_int_couple:	w_vector (w_native_int_couple,	ob, v->v.vec.p.vcnat, v->v.vec.vsize);	break;
  case TV_Vint32_couple:	w_vector (w_int32_couple,	ob, v->v.vec.p.vci32, v->v.vec.vsize);	break;
  case TV_Vint64_couple:	w_vector (w_int64_couple,	ob, v->v.vec.p.vci64, v->v.vec.vsize);	break;
  case TV_Vfloat_couple:	w_vector (w_float_couple,	ob, v->v.vec.p.vcf, v->v.vec.vsize);	break;
  case TV_Vbool_triple:		w_vector (w_bool_triple,	ob, v->v.vec.p.vtb, v->v.vec.vsize);	break;
  case TV_Vint_triple:		w_vector (w_int_triple,		ob, v->v.vec.p.vti, v->v.vec.vsize);	break;
  case TV_Vnative_int_triple:	w_vector (w_native_int_triple,	ob, v->v.vec.p.vtnat, v->v.vec.vsize);	break;
  case TV_Vint32_triple:	w_vector (w_int32_triple,	ob, v->v.vec.p.vti32, v->v.vec.vsize);	break;
  case TV_Vint64_triple:	w_vector (w_int64_triple,	ob, v->v.vec.p.vti64, v->v.vec.vsize);	break;
  case TV_Vfloat_triple:	w_vector (w_float_triple,	ob, v->v.vec.p.vtf, v->v.vec.vsize);	break;
    /* w_mat */
  case TV_Mc_float:		w_matrix (LAYOUT_C,		ob, &v->v.mat);				break;
  case TV_Mfortran_float:	w_matrix (LAYOUT_FORTRAN,	ob, &v->v.mat);				break;
  default:
    assert (0);
  }
  /* w_item */
  buf_add_string (ob, ";\n");
}

void write_typed_value (pio_buffer *ob, TV_t *v) {
  buf_add_string (ob, "begin\n");
  w_value_type (ob, type_of_value (v));
  buf_add_char (ob, '\n');
  w_value (ob, v);
  buf_add_string (ob, "end\n\n");
}
